use super::{Keyfork, KeyforkCommands};
use clap::{Parser, Subcommand};
use keyfork_mnemonic_generate::generate_mnemonic;

#[derive(Clone, Debug)]
pub enum EntropySize {
    Bits128,
    Bits256,
}

#[derive(thiserror::Error, Debug, Clone)]
pub enum EntropySizeError {
    #[error("Expected one of 128, 256")]
    InvalidChoice,
}

impl std::str::FromStr for EntropySize {
    type Err = EntropySizeError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(match s {
            "128" => EntropySize::Bits128,
            "256" => EntropySize::Bits256,
            _ => return Err(EntropySizeError::InvalidChoice),
        })
    }
}

impl From<&EntropySize> for usize {
    fn from(value: &EntropySize) -> Self {
        match value {
            EntropySize::Bits128 => 128,
            EntropySize::Bits256 => 256,
        }
    }
}

#[derive(Subcommand, Clone, Debug)]
pub enum MnemonicSubcommands {
    /// Generate a mnemonic using OS entropy.
    Generate {
        /// The size in bits to generate entropy for.
        entropy: EntropySize,
    },
}

impl MnemonicSubcommands {
    pub fn handle(
        &self,
        _m: &Mnemonic,
        _p: &KeyforkCommands,
        _keyfork: &Keyfork,
    ) -> Result<(), Box<dyn std::error::Error>> {
        match self {
            MnemonicSubcommands::Generate { entropy } => {
                let mnemonic = generate_mnemonic(usize::from(entropy))?;
                println!("{mnemonic}");
                Ok(())
            }
        }
    }
}

#[derive(Parser, Debug, Clone)]
pub struct Mnemonic {
    #[command(subcommand)]
    pub command: MnemonicSubcommands,
}
